"""Module for modeling flow cells."""
import datetime
import logging
from pathlib import Path
from typing import List, Optional, Type, Union

from pydantic.v1 import ValidationError
from typing_extensions import Literal

from cg.apps.demultiplex.sample_sheet.models import (
    FlowCellSampleNovaSeq6000Bcl2Fastq,
    FlowCellSampleNovaSeq6000Dragen,
    FlowCellSampleNovaSeqX,
    SampleSheet,
)
from cg.apps.demultiplex.sample_sheet.read_sample_sheet import get_sample_sheet_from_file
from cg.constants.constants import LENGTH_LONG_DATE
from cg.constants.demultiplexing import (
    BclConverter,
    DemultiplexingDirsAndFiles,
)
from cg.constants.sequencing import Sequencers, sequencer_types
from cg.exc import FlowCellError, SampleSheetError
from cg.models.demultiplex.run_parameters import (
    RunParameters,
    RunParametersNovaSeq6000,
    RunParametersNovaSeqX,
)

LOG = logging.getLogger(__name__)


class FlowCellDirectoryData:
    """Class to collect information about flow cell directories and their particular files."""

    def __init__(self, flow_cell_path: Path, bcl_converter: Optional[str] = "bcl2fastq"):
        LOG.debug(f"Instantiating FlowCellDirectoryData with path {flow_cell_path}")
        self.path: Path = flow_cell_path
        self.bcl_converter: Optional[str] = bcl_converter
        self._run_parameters: Optional[RunParameters] = None
        self.run_date: datetime.datetime = datetime.datetime.now()
        self.machine_name: str = ""
        self.machine_number: int = 0
        self.base_name: str = ""  # Base name is flow cell-id + flow cell position
        self.id: str = ""
        self.position: Literal["A", "B"] = "A"
        self.parse_flow_cell_dir_name()

    def parse_flow_cell_dir_name(self):
        """Parse relevant information from flow cell name.

        This will assume that the flow cell naming convention is used. If not we skip the flow cell.
        Convention is: <date>_<machine>_<run_numbers>_<A|B><flow_cell_id>
        Example: '201203_A00689_0200_AHVKJCDRXX'.
        """

        self.validate_flow_cell_dir_name()
        self.run_date = self._parse_date()
        self.machine_name = self.split_flow_cell_name[1]
        self.machine_number = int(self.split_flow_cell_name[2])
        base_name: str = self.split_flow_cell_name[-1]
        self.base_name = base_name
        LOG.debug(f"Set flow cell id to {base_name}")
        self.id = base_name[1:]
        self.position = base_name[0]

    @property
    def split_flow_cell_name(self) -> List[str]:
        """Return split flow cell name."""
        return self.path.name.split("_")

    @property
    def full_name(self) -> str:
        """Return flow cell full name."""
        return self.path.name

    @property
    def sample_sheet_path(self) -> Path:
        """
        Return sample sheet path.
        """
        return Path(self.path, DemultiplexingDirsAndFiles.SAMPLE_SHEET_FILE_NAME.value)

    @property
    def run_parameters_path(self) -> Path:
        """Return path to run parameters file."""
        return Path(self.path, DemultiplexingDirsAndFiles.RUN_PARAMETERS)

    @property
    def run_parameters(self) -> RunParameters:
        """Return run parameters object."""
        if not self.run_parameters_path.exists():
            message = f"Could not find run parameters file {self.run_parameters_path}"
            LOG.warning(message)
            raise FileNotFoundError(message)
        if not self._run_parameters:
            self._run_parameters = (
                RunParametersNovaSeqX(run_parameters_path=self.run_parameters_path)
                if self.sequencer_type == Sequencers.NOVASEQX
                else RunParametersNovaSeq6000(run_parameters_path=self.run_parameters_path)
            )
        return self._run_parameters

    @property
    def sample_type(
        self,
    ) -> Union[
        Type[FlowCellSampleNovaSeq6000Bcl2Fastq],
        Type[FlowCellSampleNovaSeq6000Dragen],
        Type[FlowCellSampleNovaSeqX],
    ]:
        """Return the sample class used in the flow cell."""
        if self.sequencer_type == Sequencers.NOVASEQX:
            return FlowCellSampleNovaSeqX
        if (
            self.bcl_converter == BclConverter.DRAGEN
            or self.bcl_converter == BclConverter.BCLCONVERT
        ):
            return FlowCellSampleNovaSeq6000Dragen
        return FlowCellSampleNovaSeq6000Bcl2Fastq

    @property
    def sequencer_type(
        self,
    ) -> Literal[Sequencers.HISEQX, Sequencers.HISEQGA, Sequencers.NOVASEQ, Sequencers.NOVASEQX]:
        """Return the sequencer type."""
        return sequencer_types[self.machine_name]

    @property
    def rta_complete_path(self) -> Path:
        """Return RTAComplete path."""
        return Path(self.path, DemultiplexingDirsAndFiles.RTACOMPLETE)

    @property
    def copy_complete_path(self) -> Path:
        """Return copy complete path."""
        return Path(self.path, DemultiplexingDirsAndFiles.COPY_COMPLETE)

    @property
    def hiseq_x_copy_complete_path(self) -> Path:
        """Return copy complete path for Hiseq X."""
        return Path(self.path, DemultiplexingDirsAndFiles.Hiseq_X_COPY_COMPLETE)

    @property
    def hiseq_x_delivery_started_path(self) -> Path:
        """Return delivery started path for Hiseq X."""
        return Path(self.path, DemultiplexingDirsAndFiles.DELIVERY)

    @property
    def demultiplexing_started_path(self) -> Path:
        """Return demux started path."""
        return Path(self.path, DemultiplexingDirsAndFiles.DEMUX_STARTED)

    @property
    def trailblazer_config_path(self) -> Path:
        """Return file to SLURM job ids path."""
        return Path(self.path, "slurm_job_ids.yaml")

    @property
    def hiseq_x_flow_cell(self) -> Path:
        """Return path to Hiseq X flow cell directory."""
        return Path(self.path, DemultiplexingDirsAndFiles.Hiseq_X_TILE_DIR)

    @property
    def is_demultiplexing_complete(self) -> bool:
        return Path(self.path, DemultiplexingDirsAndFiles.DEMUX_COMPLETE).exists()

    def _parse_date(self):
        """Return the parsed date in the correct format."""
        if len(self.split_flow_cell_name[0]) == LENGTH_LONG_DATE:
            return datetime.datetime.strptime(self.split_flow_cell_name[0], "%Y%m%d")
        return datetime.datetime.strptime(self.split_flow_cell_name[0], "%y%m%d")

    def validate_flow_cell_dir_name(self) -> None:
        """
        Validate on the following criteria:
        Convention is: <date>_<machine>_<run_numbers>_<A|B><flow_cell_id>
        Example: '201203_A00689_0200_AHVKJCDRXX'.
        """
        if len(self.split_flow_cell_name) != 4:
            message = f"Flowcell {self.full_name} does not follow the flow cell naming convention"
            LOG.warning(message)
            raise FlowCellError(message)

    def is_demultiplexing_started(self) -> bool:
        """Check if demultiplexing started path exists."""
        return self.demultiplexing_started_path.exists()

    def sample_sheet_exists(self) -> bool:
        """Check if sample sheet exists."""
        LOG.info("Check if sample sheet exists")
        return self.sample_sheet_path.exists()

    def validate_sample_sheet(self) -> bool:
        """Validate if sample sheet is on correct format."""
        try:
            get_sample_sheet_from_file(
                infile=self.sample_sheet_path,
                flow_cell_sample_type=self.sample_type,
            )
        except (SampleSheetError, ValidationError) as error:
            LOG.warning("Invalid sample sheet")
            LOG.warning(error)
            return False
        return True

    def get_sample_sheet(self) -> SampleSheet:
        """Return sample sheet object."""
        return get_sample_sheet_from_file(
            infile=self.sample_sheet_path,
            flow_cell_sample_type=self.sample_type,
        )

    def is_sequencing_done(self) -> bool:
        """Check if sequencing is done.
        This is indicated by that the file RTAComplete.txt exists.
        """
        LOG.info("Check if sequencing is done")
        return self.rta_complete_path.exists()

    def is_copy_completed(self) -> bool:
        """Check if copy of flow cell is done.
        This is indicated by that the file CopyComplete.txt exists.
        """
        LOG.info("Check if copy of data from sequence instrument is ready")
        return self.copy_complete_path.exists()

    def is_hiseq_x_copy_completed(self) -> bool:
        """Check if copy of Hiseq X flow cell is done."""
        LOG.info("Check if copy of data from Hiseq X sequence instrument is ready")
        return self.hiseq_x_copy_complete_path.exists()

    def is_hiseq_x_delivery_started(self) -> bool:
        """Check if delivery of Hiseq X flow cell is started."""
        LOG.info("Check if delivery of data from Hiseq X sequence instrument is ready")
        return self.hiseq_x_delivery_started_path.exists()

    def is_hiseq_x(self) -> bool:
        """Check if flow cell is Hiseq X."""
        LOG.debug("Check if flow cell is Hiseq X")
        return self.hiseq_x_flow_cell.exists()

    def is_flow_cell_ready(self) -> bool:
        """Check if a flow cell is ready for demultiplexing.

        A flow cell is ready if the two files RTAComplete.txt and CopyComplete.txt exists in the
        flow cell directory.
        """
        LOG.info("Check if flow cell is ready for demultiplexing")
        if not self.is_sequencing_done():
            LOG.info(f"Sequencing is not completed for flow cell {self.id}")
            return False
        LOG.debug(f"Sequence is done for flow cell {self.id}")
        if not self.is_copy_completed():
            LOG.info(f"Copy of sequence data is not ready for flow cell {self.id}")
            return False
        LOG.debug(f"All data has been transferred for flow cell {self.id}")
        LOG.info(f"Flow cell {self.id} is ready for demultiplexing")
        return True

    def __str__(self):
        return f"FlowCell(path={self.path},run_parameters_path={self.run_parameters_path})"



"""Models used by hermes <-> cg interactions"""
import logging
from pathlib import Path
from typing import List, Optional

from pydantic.v1 import BaseModel, validator

LOG = logging.getLogger(__name__)


class CGTag(BaseModel):
    """Specify the file output information from hermes"""

    path: str
    tags: List[str]
    mandatory: Optional[bool] = False


class CGDeliverables(BaseModel):
    """Class that specifies the output from hermes"""

    pipeline: str
    bundle_id: str
    files: List[CGTag]

    @validator("files", each_item=True)
    def check_mandatory_exists(cls, file):
        if file.mandatory:
            assert Path(file.path).exists(), f"Mandatory file cannot be found at {file.path}"
            return file
        if not Path(file.path).exists():
            LOG.info("Optional file %s not found, removing from bundle", file.path)
            return
        return file

    @validator("files")
    def remove_invalid(cls, value):
        return [item for item in value if item]



from sqlalchemy import Column, orm, types

from .base import Model


class Project(Model):
    project_id = Column(types.Integer, primary_key=True)
    projectname = Column(types.String(255), nullable=False)
    comment = Column(types.Text)
    time = Column(types.DateTime, nullable=False)

    samples = orm.relationship("Sample", cascade="all", backref=orm.backref("project"))

    def __repr__(self):
        return "{self.__class__.__name__}: {self.project_id}".format(self=self)



import itertools
import logging
import shutil

from glob import glob
from pathlib import Path
from typing import Iterable, List, Optional
from cg.apps.demultiplex.demultiplex_api import DemultiplexingAPI
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.cgstats.stats import StatsAPI
from cg.constants import SequencingFileTag
from cg.exc import DeleteDemuxError
from cg.models.cg_config import CGConfig
from cg.store import Store
from cg.store.models import Sample, Flowcell
from housekeeper.store.models import File

LOG = logging.getLogger(__name__)


class DeleteDemuxAPI:
    """Class to handle wiping out a flow cell before restart/start"""

    def __init__(self, config: CGConfig, dry_run: bool, flow_cell_name: str):
        self.dry_run: bool = self.set_dry_run(dry_run=dry_run)
        self.flow_cell_name = flow_cell_name
        self.housekeeper_api: HousekeeperAPI = config.housekeeper_api
        self.status_db: Store = config.status_db
        self.demux_api: DemultiplexingAPI = config.demultiplex_api
        self.stats_api: StatsAPI = config.cg_stats_api
        self.samples_on_flow_cell: List[Sample] = []
        self.demultiplexing_out_path: Path = self.get_path_for_flow_cell(
            base_path=self.demux_api.out_dir
        )
        self.run_path: Path = self.get_path_for_flow_cell(base_path=self.demux_api.run_dir)
        LOG.debug("DeleteDemuxAPI: API initiated")

    def get_path_for_flow_cell(
        self,
        base_path: Path,
    ) -> Path:
        """
        Return the path to the run dir or demultiplex dir for the given flow cell name
        depending on the base_path input.
        """
        flow_cell_path = next(base_path.rglob(pattern=f"*{self.flow_cell_name}"), None)
        if flow_cell_path and flow_cell_path.exists():
            return flow_cell_path
        else:
            LOG.warning(
                f"DeleteDemuxAPI: Path {flow_cell_path} not found for {self.flow_cell_name}"
            )

    @property
    def status_db_presence(self) -> bool:
        """Update about the presence of given flow cell in status_db"""
        return bool(self.status_db.get_flow_cell_by_name(flow_cell_name=self.flow_cell_name))

    @staticmethod
    def set_dry_run(dry_run: bool) -> bool:
        """Set dry run flag for API"""
        LOG.debug(f"DeleteDemuxAPI: Setting dry run mode to {dry_run}")
        return dry_run

    def _set_samples_on_flow_cell(self) -> None:
        """Set a list of samples related to a flow cell in status-db."""
        flow_cell = self.status_db.get_flow_cell_by_name(flow_cell_name=self.flow_cell_name)
        self.samples_on_flow_cell: List[Sample] = flow_cell.samples

    def active_samples_on_flow_cell(self) -> Optional[List[str]]:
        """Check if there are any active cases related to samples of a flow cell."""
        active_samples_on_flow_cell: List[str] = [
            sample.internal_id
            for sample in self.samples_on_flow_cell
            if self.status_db.has_active_cases_for_sample(internal_id=sample.internal_id)
        ]
        if active_samples_on_flow_cell:
            return active_samples_on_flow_cell

    def _delete_sample_sheet_housekeeper(self) -> None:
        """Delete the presence of all sample sheets related to a flow cell in Housekeeper."""
        sample_sheet_files: Iterable[File] = self.housekeeper_api.files(
            tags=[self.flow_cell_name, "samplesheet"]
        )
        if any(sample_sheet_files):
            for file in sample_sheet_files:
                self.housekeeper_api.delete_file(file_id=file.id)
                LOG.info(f"DeleteDemuxAPI-Housekeeper: Deleted {file.path} from housekeeper")
        else:
            LOG.info(f"DeleteDemuxAPI-Housekeeper: No files found with tag: {self.flow_cell_name}")

    def _delete_files_if_related_in_housekeeper_by_tag(self, sample: Sample, tags: List[str]):
        """Delete any existing fastq related to sample"""

        housekeeper_files: Iterable[File] = self.housekeeper_api.files(
            bundle=sample.internal_id, tags=tags
        )
        if not housekeeper_files:
            LOG.info(f"Could not find {tags} for {sample.internal_id}")
        else:
            for housekeeper_file in housekeeper_files:
                self.housekeeper_api.delete_file(file_id=housekeeper_file.id)

    def _delete_fastq_and_spring_housekeeper(self) -> None:
        """Delete the presence of any spring/fastq files in Housekeeper related to samples on the flow cell."""

        tag_combinations: List[List[str]] = [
            [SequencingFileTag.FASTQ.value, self.flow_cell_name],
            [SequencingFileTag.SPRING.value, self.flow_cell_name],
            [SequencingFileTag.SPRING_METADATA.value, self.flow_cell_name],
        ]
        for tags, sample in itertools.product(tag_combinations, self.samples_on_flow_cell):
            self._delete_files_if_related_in_housekeeper_by_tag(sample=sample, tags=tags)

    def delete_flow_cell_housekeeper(self) -> None:
        """Delete any presence of a flow cell in housekeeper. Including Sample sheets AND fastq-files"""

        if self.dry_run:
            LOG.info(
                f"DeleteDemuxAPI-Housekeeper: Would delete sample sheet files with tag {self.flow_cell_name}"
            )
            LOG.info(
                f"DeleteDemuxAPI-Housekeeper: Would delete fastq and spring files related to flow cell {self.flow_cell_name}"
            )
        else:
            LOG.info(
                f"DeleteDemuxAPI-Housekeeper: Deleting sample sheet files with tag {self.flow_cell_name}"
            )
            self._delete_sample_sheet_housekeeper()
            LOG.info(
                f"DeleteDemuxAPI-Housekeeper: Deleting fastq and spring files related to flowcell {self.flow_cell_name}"
            )
            self._delete_fastq_and_spring_housekeeper()

    def delete_flow_cell_in_status_db(self) -> None:
        """Delete a flow cell in Status db."""
        if self.dry_run:
            LOG.info(f"DeleteDemuxAPI-StatusDB: Would remove {self.flow_cell_name}")
        else:
            self.status_db.delete_flow_cell(flow_cell_id=self.flow_cell_name)
            LOG.info(f"DeleteDemuxAPI-StatusDB: Deleted flowcell {self.flow_cell_name}")

    def delete_flow_cell_cgstats(self) -> None:
        """Delete any presence of a flow cell in cgstats"""
        from cg.apps.cgstats.crud.delete import delete_flowcell

        if self.dry_run:
            LOG.info(f"DeleteDemuxAPI-CGStats: Would remove {self.flow_cell_name}")
        else:
            delete_flowcell(manager=self.stats_api, flowcell_name=self.flow_cell_name)

    def _delete_demultiplexing_dir_hasta(self, demultiplexing_dir: bool) -> None:
        """delete demultiplexing directory on server."""
        if demultiplexing_dir and self.demultiplexing_out_path.exists():
            LOG.info(
                f"DeleteDemuxAPI-Hasta: Removing flow cell demultiplexing directory: {self.demultiplexing_out_path}"
            )
            shutil.rmtree(self.demultiplexing_out_path, ignore_errors=False)
        else:
            LOG.info(
                f"DeleteDemuxAPI-Hasta: No demultiplexing directory found for {self.flow_cell_name}"
            )

    def _delete_run_dir_hasta(self, run_dir: bool) -> None:
        """delete flow cell run directory on server."""
        if run_dir and self.run_path.exists():
            LOG.info(f"DeleteDemuxAPI-Hasta: Removing flow cell run directory: {self.run_path}")
            shutil.rmtree(path=self.run_path, ignore_errors=False)
        else:
            LOG.info(f"DeleteDemuxAPI-Hasta: No run directory found for {self.flow_cell_name}")

    def delete_flow_cell_hasta(
        self,
        demultiplexing_dir: bool,
        run_dir: bool,
    ) -> None:
        """Delete a flow cells presence on the server, if flow cell is removed from demultiplexing dir and run
        dir set status to removed."""
        if self.dry_run:
            LOG.info(
                f"DeleteDemuxAPI-Hasta: Would have removed the following directory: {self.demultiplexing_out_path}\n"
                f"DeleteDemuxAPI-Hasta: Would have removed the following directory: {self.run_path}"
            )
            return
        self.set_flow_cell_status_statusdb_to_removed(
            demultiplexing_dir=demultiplexing_dir,
            run_dir=run_dir,
            status_db=self.status_db_presence,
        )
        self._delete_demultiplexing_dir_hasta(demultiplexing_dir=demultiplexing_dir)
        self._delete_run_dir_hasta(run_dir=run_dir)

    def set_flow_cell_status_statusdb_to_removed(
        self, demultiplexing_dir: bool, run_dir: bool, status_db: bool
    ) -> None:
        if demultiplexing_dir and run_dir and status_db:
            flow_cell_obj: Flowcell = self.status_db.get_flow_cell_by_name(
                flow_cell_name=self.flow_cell_name
            )
            flow_cell_obj.status = "removed"
            self.status_db.session.commit()

    def delete_demux_init_files(self):
        """Delete previous traces of slurm job ids."""
        slurm_job_id_file_path: Path = Path(self.run_path, "slurm_job_ids.yaml")
        demux_script_file_path: Path = Path(self.run_path, "demux-novaseq.sh")
        try:
            error_log_path, log_path = glob(
                f"{self.run_path}/{self.flow_cell_name}_demultiplex.std*"
            )
            demux_init_files: List[Path] = [
                slurm_job_id_file_path,
                demux_script_file_path,
                Path(error_log_path),
                Path(log_path),
            ]
        except ValueError:
            LOG.info(f"DeleteDemuxAPI-Init-files: No demultiplexing logs found in: {self.run_path}")
            demux_init_files: List[Path] = [
                slurm_job_id_file_path,
                demux_script_file_path,
            ]
        if self.dry_run:
            LOG.info(f"DeleteDemuxAPI-Init-files: Would have removed {demux_init_files}")
        for init_file in demux_init_files:
            if init_file.is_file():
                LOG.info(f"DeleteDemuxAPI: Removing {init_file}")
                init_file.unlink()
            else:
                LOG.info(f"DeleteDemuxAPI: Did not find {init_file}")

    def check_active_samples(self) -> None:
        """Check, and raise, if there are active samples on a flow cell."""
        if self.status_db_presence:
            self._set_samples_on_flow_cell()
            active_samples_on_flow_cell: List[str] = self.active_samples_on_flow_cell()
            if active_samples_on_flow_cell:
                LOG.warning(
                    f"There are active analyses using data from this flowcell.\n"
                    f"Before restarting the demultiplexing process - please cancel the ongoing analyses of the "
                    f"following sample(s): \n{active_samples_on_flow_cell}"
                )
                raise DeleteDemuxError

    def delete_flow_cell(
        self,
        cg_stats: bool,
        demultiplexing_dir: bool,
        run_dir: bool,
        housekeeper: bool,
        init_files: bool,
        status_db: bool,
    ) -> None:
        """Master command to delete the presence of a flow cell in all services."""
        self.check_active_samples()
        if status_db:
            self.delete_flow_cell_hasta(
                demultiplexing_dir=True,
                run_dir=True,
            )
            self.delete_flow_cell_cgstats()
            self.delete_flow_cell_housekeeper()
            self.delete_flow_cell_in_status_db()
        if demultiplexing_dir or run_dir:
            self.delete_flow_cell_hasta(
                demultiplexing_dir=demultiplexing_dir,
                run_dir=run_dir,
            )
        if cg_stats:
            self.delete_flow_cell_cgstats()
        if init_files and not run_dir:
            self.delete_demux_init_files()
        if housekeeper:
            self.delete_flow_cell_housekeeper()



"""Tests for transfer flow cell data."""
from datetime import datetime
import logging
import warnings
from pathlib import Path
from typing import Generator

from sqlalchemy import exc as sa_exc

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants import FlowCellStatus, SequencingFileTag
from cg.meta.transfer import TransferFlowCell
from cg.meta.transfer.flowcell import _set_status_db_sample_sequenced_at, log_enough_reads
from cg.models.cgstats.flowcell import StatsFlowcell
from cg.models.demultiplex.demux_results import DemuxResults
from cg.store import Store
from cg.store.models import Flowcell, Sample
from tests.store_helpers import StoreHelpers


def test_add_tags_to_housekeeper(
    bcl2fastq_flow_cell_id: str, transfer_flow_cell_api: Generator[TransferFlowCell, None, None]
):
    """Test adding tag to Housekeeper."""
    # GIVEN transfer flow cell API

    # GIVEN no flow cell id tag in Housekeeper
    assert transfer_flow_cell_api.hk.get_tag(name=bcl2fastq_flow_cell_id) is None

    # WHEN adding tags to Housekeeper
    transfer_flow_cell_api._add_tags_to_housekeeper(
        store=True,
        tags=[SequencingFileTag.FASTQ, SequencingFileTag.SAMPLE_SHEET, bcl2fastq_flow_cell_id],
    )

    # THEN tha tags should be added
    assert transfer_flow_cell_api.hk.get_tag(name=bcl2fastq_flow_cell_id) is not None


def test_add_flow_cell_to_status_db(
    transfer_flow_cell_api: Generator[TransferFlowCell, None, None], bcl2fastq_flow_cell_id: str
):
    """Test adding flow cell to Status db."""
    # GIVEN transfer flow cell API

    # GIVEN a flow cell that does not exist in status db
    flow_cell: Flowcell = transfer_flow_cell_api.db.get_flow_cell_by_name(
        flow_cell_name=bcl2fastq_flow_cell_id
    )

    assert flow_cell is None

    # GIVEN a cgstats flow cell
    cgstats_flow_cell: StatsFlowcell = transfer_flow_cell_api.stats.flowcell(bcl2fastq_flow_cell_id)

    # WHEN adding flow cell to status db
    added_flow_cell: Flowcell = transfer_flow_cell_api._add_flow_cell_to_status_db(
        cgstats_flow_cell=cgstats_flow_cell,
        flow_cell=flow_cell,
        flow_cell_id=bcl2fastq_flow_cell_id,
    )

    # THEN the flow cell added/returned should have the same id
    assert added_flow_cell.name == bcl2fastq_flow_cell_id


def test_add_flow_cell_to_status_db_existing_flow_cell(
    flowcell_store: Store,
    bcl2fastq_flow_cell_id: str,
    helpers: StoreHelpers,
    transfer_flow_cell_api: Generator[TransferFlowCell, None, None],
):
    """Test adding flow cell to Status db, when already present."""
    # GIVEN transfer flow cell API

    # GIVEN a flow cell that exist in status db
    helpers.add_flowcell(store=flowcell_store, flow_cell_name=bcl2fastq_flow_cell_id)
    flow_cell: Flowcell = transfer_flow_cell_api.db.get_flow_cell_by_name(
        flow_cell_name=bcl2fastq_flow_cell_id
    )

    assert flow_cell is not None

    # GIVEN a cgstats flow cell
    cgstats_flow_cell: StatsFlowcell = transfer_flow_cell_api.stats.flowcell(bcl2fastq_flow_cell_id)

    # WHEN adding flow cell to status db
    added_flow_cell: Flowcell = transfer_flow_cell_api._add_flow_cell_to_status_db(
        cgstats_flow_cell=cgstats_flow_cell,
        flow_cell=flow_cell,
        flow_cell_id=bcl2fastq_flow_cell_id,
    )

    # THEN flow cell should be returned should be identical to the one supplied
    assert added_flow_cell is flow_cell


def test_include_sample_sheet_to_housekeeper_when_not_existing(
    caplog,
    bcl2fastq_flow_cell_id: str,
    transfer_flow_cell_api: Generator[TransferFlowCell, None, None],
):
    """Test including sample sheet to Housekeeper when none can be found."""
    # GIVEN transfer flow cell API

    # GIVEN a sample sheet tag in Housekeeper
    transfer_flow_cell_api._add_tags_to_housekeeper(
        store=True, tags=[SequencingFileTag.SAMPLE_SHEET]
    )

    # WHEN including sample sheet to Housekeeper
    transfer_flow_cell_api._include_sample_sheet_to_housekeeper(
        flow_cell_dir=Path("does_not_exist"), flow_cell_id=bcl2fastq_flow_cell_id, store=True
    )

    # THEN tha sample sheet should not be found
    assert "Unable to find file:" in caplog.text


def test_include_sample_sheet_to_housekeeper(
    bcl2fastq_flow_cell_id: str,
    sample_sheet_path: Generator[Path, None, None],
    transfer_flow_cell_api: Generator[TransferFlowCell, None, None],
):
    """Test including sample sheet to Housekeeper."""
    # GIVEN transfer flow cell API

    # GIVEN a sample sheet tag in Housekeeper
    transfer_flow_cell_api._add_tags_to_housekeeper(
        store=True, tags=[SequencingFileTag.SAMPLE_SHEET]
    )

    # GIVEN no flow cell id bundle in housekeeper
    hk_bundle = transfer_flow_cell_api.hk.bundle(name=SequencingFileTag.SAMPLE_SHEET)
    assert hk_bundle is None

    # GIVEN a sample sheet that exists

    # WHEN including sample sheet to Housekeeper
    transfer_flow_cell_api._include_sample_sheet_to_housekeeper(
        flow_cell_dir=sample_sheet_path.parent, flow_cell_id=bcl2fastq_flow_cell_id, store=True
    )

    # THEN the sample sheet should be included to Housekeeper
    hk_bundle = transfer_flow_cell_api.hk.bundle(name=SequencingFileTag.SAMPLE_SHEET)
    for hk_file in hk_bundle.versions[0].files:
        assert hk_file.path.endswith("csv")


def test_include_cgstats_log_to_housekeeper_when_not_existing(
    bcl2fastq_flow_cell_id: str,
    transfer_flow_cell_api: Generator[TransferFlowCell, None, None],
):
    """Test including cgstats log to Housekeeper when not existing."""
    # GIVEN transfer flow cell API

    # GIVEN a cgstats log tag in Housekeeper
    transfer_flow_cell_api._add_tags_to_housekeeper(
        store=True, tags=[SequencingFileTag.CGSTATS_LOG]
    )

    # GIVEN no flow cell id bundle in housekeeper
    hk_bundle = transfer_flow_cell_api.hk.bundle(name=SequencingFileTag.CGSTATS_LOG)
    assert hk_bundle is None

    # GIVEN no cgstats log file

    # WHEN including cgstats log file to Housekeeper
    transfer_flow_cell_api._include_cgstats_log_to_housekeeper(
        flow_cell_dir=Path("does_not_exist"), flow_cell_id=bcl2fastq_flow_cell_id, store=True
    )

    # THEN the cgstats log file not be included to Housekeeper
    hk_bundle = transfer_flow_cell_api.hk.bundle(name=SequencingFileTag.CGSTATS_LOG)
    assert hk_bundle is None


def test_include_cgstats_log_to_housekeeper(
    bcl2fastq_flow_cell_id: str,
    cgstats_log_path: Generator[Path, None, None],
    transfer_flow_cell_api: Generator[TransferFlowCell, None, None],
):
    """Test including cgstats log to Housekeeper."""
    # GIVEN transfer flow cell API

    # GIVEN a cgstats log tag in Housekeeper
    transfer_flow_cell_api._add_tags_to_housekeeper(
        store=True, tags=[SequencingFileTag.CGSTATS_LOG]
    )

    # GIVEN no flow cell id bundle in housekeeper
    hk_bundle = transfer_flow_cell_api.hk.bundle(name=SequencingFileTag.CGSTATS_LOG)
    assert hk_bundle is None

    # GIVEN a cgstats log file that exists

    # WHEN including cgstats log file to Housekeeper
    transfer_flow_cell_api._include_cgstats_log_to_housekeeper(
        flow_cell_dir=cgstats_log_path.parent, flow_cell_id=bcl2fastq_flow_cell_id, store=True
    )

    # THEN the cgstats log file be included to Housekeeper
    hk_bundle = transfer_flow_cell_api.hk.bundle(name=SequencingFileTag.CGSTATS_LOG)
    for hk_file in hk_bundle.versions[0].files:
        assert hk_file.path.endswith("txt")


def test_store_sequencing_files(
    caplog,
    sample_sheet_path: Generator[Path, None, None],
    bcl2fastq_flow_cell_id: str,
    transfer_flow_cell_api: Generator[TransferFlowCell, None, None],
):
    """Test storing sequencing files to Housekeeper."""
    caplog.set_level(logging.INFO)
    # GIVEN transfer flow cell API

    # GIVEN a sample sheet tag in Housekeeper
    transfer_flow_cell_api._add_tags_to_housekeeper(
        store=True, tags=[SequencingFileTag.SAMPLE_SHEET]
    )

    # GIVEN no flow cell id bundle in housekeeper
    hk_bundle = transfer_flow_cell_api.hk.bundle(name=SequencingFileTag.SAMPLE_SHEET)
    assert hk_bundle is None

    # GIVEN a sample sheet that exists

    # WHEN adding sample sheet to Housekeeper
    transfer_flow_cell_api._store_sequencing_files(
        flow_cell_id=bcl2fastq_flow_cell_id,
        sequencing_files=[sample_sheet_path.as_posix()],
        tag_name=SequencingFileTag.SAMPLE_SHEET,
    )

    # THEN we should log that we are adding a file
    assert f"Adding file using tag: {SequencingFileTag.SAMPLE_SHEET}" in caplog.text

    # THEN tha sample sheet should be added to Housekeeper
    hk_bundle = transfer_flow_cell_api.hk.bundle(name=SequencingFileTag.SAMPLE_SHEET)
    for hk_file in hk_bundle.versions[0].files:
        assert hk_file.path.endswith("csv")


def test_include_sequencing_file(
    caplog,
    bcl2fastq_flow_cell_id: str,
    sample_sheet_path: Generator[Path, None, None],
    transfer_flow_cell_api: Generator[TransferFlowCell, None, None],
):
    """Test including sequencing files to Housekeeper."""
    caplog.set_level(logging.INFO)
    # GIVEN transfer flow cell API

    # GIVEN a sample sheet tag in Housekeeper
    transfer_flow_cell_api._add_tags_to_housekeeper(
        store=True, tags=[SequencingFileTag.SAMPLE_SHEET]
    )

    # GIVEN no flow cell id bundle in housekeeper
    hk_bundle = transfer_flow_cell_api.hk.bundle(name=SequencingFileTag.SAMPLE_SHEET)
    assert hk_bundle is None

    # GIVEN a sample sheet that exists

    # WHEN adding sample sheet to Housekeeper
    transfer_flow_cell_api._store_sequencing_files(
        flow_cell_id=bcl2fastq_flow_cell_id,
        sequencing_files=[sample_sheet_path.as_posix()],
        tag_name=SequencingFileTag.SAMPLE_SHEET,
    )

    # THEN we should log that we are adding a file
    assert f"Adding file using tag: {SequencingFileTag.SAMPLE_SHEET}" in caplog.text

    # THEN tha sample sheet should be added to Housekeeper
    hk_bundle = transfer_flow_cell_api.hk.bundle(name=SequencingFileTag.SAMPLE_SHEET)
    for hk_file in hk_bundle.versions[0].files:
        assert hk_file.path.endswith("csv")


def test_set_status_db_sample_sequenced_at_when_first_sequenced(
    base_store: Store, helpers: StoreHelpers, timestamp_now: datetime
):
    """Test setting sample sequenced at with no previous sequencing."""
    # GIVEN a status db sample
    sample: Sample = helpers.add_sample(base_store, sequenced_at=None)

    # Given no previous sequencing
    assert sample.sequenced_at is None

    # WHEN setting sequenced at for the sample
    _set_status_db_sample_sequenced_at(
        status_db_sample=sample, flow_cell_sequenced_at=timestamp_now
    )

    # THEN the sample sequenced at should be set
    assert sample.sequenced_at == timestamp_now


def test_set_status_db_sample_sequenced_at_when_sequenced_again(
    base_store: Store,
    helpers: StoreHelpers,
    timestamp_now: datetime,
    timestamp_yesterday: datetime,
):
    """Test setting sample sequenced at when sequenced again."""
    # GIVEN a status db sample sequenced yesterday
    sample: Sample = helpers.add_sample(base_store, sequenced_at=timestamp_yesterday)

    # WHEN setting sequenced at for the sample
    _set_status_db_sample_sequenced_at(
        status_db_sample=sample, flow_cell_sequenced_at=timestamp_now
    )

    # THEN the sample sequenced at should be set to today
    assert sample.sequenced_at == timestamp_now


def test_log_enough_reads_when_enough_reads(caplog, sample_name: str):
    """Test logging enough reads."""
    caplog.set_level(logging.INFO)

    # GIVEN a status db sample reads with enough reads

    # GIVEN a sample application expected reads

    # WHEN setting sequenced at for the sample
    log_enough_reads(
        status_db_sample_reads=2, application_expected_reads=1, cgstats_sample_name=sample_name
    )

    # THEN we should log DONE
    assert f"[DONE]" in caplog.text


def test_log_enough_reads_when_not_enough_reads(caplog, sample_name: str):
    """Test logging not enough reads."""
    caplog.set_level(logging.INFO)

    # GIVEN a status db sample with not enough reads

    # GIVEN a sample application expected reads

    # WHEN setting sequenced at for the sample
    log_enough_reads(
        status_db_sample_reads=1, application_expected_reads=2, cgstats_sample_name=sample_name
    )

    # THEN the should log NOT DONE
    assert f"[NOT DONE]" in caplog.text


def test_parse_flow_cell_samples(
    flowcell_store: Store,
    helpers: StoreHelpers,
    transfer_flow_cell_api: Generator[TransferFlowCell, None, None],
    bcl2fastq_flow_cell_id: str,
):
    """Test parsing of flow cell samples."""

    # GIVEN a cgstats flow cell
    cgstats_flow_cell: StatsFlowcell = transfer_flow_cell_api.stats.flowcell(bcl2fastq_flow_cell_id)

    # GIVEN a flow cell that exist in status db
    flow_cell: Flowcell = helpers.add_flowcell(
        store=flowcell_store, flow_cell_name=bcl2fastq_flow_cell_id
    )

    # GIVEN no sample in flow cell
    assert len(flow_cell.samples) == 0

    # WHEN parsing the flow cell samples
    transfer_flow_cell_api._parse_flow_cell_samples(
        cgstats_flow_cell=cgstats_flow_cell,
        flow_cell=flow_cell,
        flow_cell_id=bcl2fastq_flow_cell_id,
        store=True,
    )

    # THEN a sample should have been added
    assert len(flow_cell.samples) == 1


def test_parse_flow_cell_samples_when_no_cgstats_sample(
    caplog,
    flowcell_store: Store,
    helpers: StoreHelpers,
    transfer_flow_cell_api: Generator[TransferFlowCell, None, None],
    bcl2fastq_flow_cell_id: str,
):
    """Test parsing of flow cell samples when no cgstats sample."""

    # GIVEN a cgstats flow cell
    cgstats_flow_cell: StatsFlowcell = transfer_flow_cell_api.stats.flowcell(bcl2fastq_flow_cell_id)

    # GIVEN a sample name that does not exist in cgstats
    cgstats_flow_cell.samples[0].name = "sample_does_not_exist_in_cgstats"

    # GIVEN a flow cell that exist in status db
    flow_cell: Flowcell = helpers.add_flowcell(
        store=flowcell_store, flow_cell_name=bcl2fastq_flow_cell_id
    )

    # GIVEN no sample in flow cell
    assert len(flow_cell.samples) == 0

    # WHEN parsing the flow cell samples
    transfer_flow_cell_api._parse_flow_cell_samples(
        cgstats_flow_cell=cgstats_flow_cell,
        flow_cell=flow_cell,
        flow_cell_id=bcl2fastq_flow_cell_id,
        store=True,
    )

    # THEN no sample should have been added
    assert len(flow_cell.samples) == 0

    # THEN we should llog that sample cannot be found
    assert f"Unable to find sample: {cgstats_flow_cell.samples[0].name}" in caplog.text


def test_transfer(
    bcl2fastq_demux_results: DemuxResults,
    flowcell_store: Store,
    transfer_flow_cell_api: Generator[TransferFlowCell, None, None],
    bcl2fastq_flow_cell_id: str,
):
    """Test transfer of sequencing files."""

    # GIVEN a store with a received but not sequenced sample
    housekeeper_api: HousekeeperAPI = transfer_flow_cell_api.hk
    assert len(flowcell_store._get_query(table=Sample).all()) == 2
    assert flowcell_store._get_query(table=Flowcell).count() == 0
    assert housekeeper_api.bundles().count() == 0

    # GIVEN a sample sheet

    # WHEN transferring the flowcell containing the sample
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=sa_exc.SAWarning)
        flow_cell = transfer_flow_cell_api.transfer(
            flow_cell_dir=bcl2fastq_demux_results.flow_cell.path,
            flow_cell_id=bcl2fastq_flow_cell_id,
        )

    # THEN it should create a new flow cell record
    assert flowcell_store._get_query(table=Flowcell).count() == 1
    assert flow_cell.status == FlowCellStatus.ON_DISK
    assert isinstance(flow_cell.id, int)
    assert flow_cell.name == bcl2fastq_flow_cell_id
    status_sample = flowcell_store._get_query(table=Sample).first()
    assert isinstance(status_sample.sequenced_at, datetime)

    # ... and it should store the fastq files and samplesheet for the sample in housekeeper
    hk_bundle = housekeeper_api.bundle(name=status_sample.internal_id)

    assert len(hk_bundle.versions[0].files) > 0
    assert (
        len([hk_file for hk_file in hk_bundle.versions[0].files if hk_file.path.endswith("csv")])
        == 1
    )

    for hk_file in hk_bundle.versions[0].files:
        assert hk_file.path.endswith("fastq.gz") or hk_file.path.endswith("csv")



from pathlib import Path
from mock import MagicMock, PropertyMock, patch

import pytest

from cg.constants.constants import FileExtensions
from cg.constants.demultiplexing import BclConverter, DemultiplexingDirsAndFiles
from cg.constants.sequencing import FLOWCELL_Q30_THRESHOLD, Sequencers
from cg.exc import FlowCellError
from cg.meta.demultiplex.utils import (
    create_delivery_file_in_flow_cell_directory,
    get_bcl_converter_name,
    get_lane_from_sample_fastq,
    get_q30_threshold,
    get_sample_sheet_path,
    parse_flow_cell_directory_data,
    copy_sample_sheet,
)
from cg.meta.demultiplex.validation import is_bcl2fastq_demux_folder_structure
from cg.models.demultiplex.flow_cell import FlowCellDirectoryData
from cg.meta.demultiplex.demux_post_processing import (
    DemuxPostProcessingAPI,
)
from cg.models.cg_config import CGConfig


def test_get_lane_from_sample_fastq_file_path():
    # GIVEN a sample fastq path
    lane = 4
    sample_fastq_path = Path(
        f"H5CYFDSX7_ACC12164A17_S367_L00{lane}_R1_001{FileExtensions.FASTQ}{FileExtensions.GZIP}",
    )

    # WHEN we get lane from the sample fastq file path
    result = get_lane_from_sample_fastq(sample_fastq_path)

    # THEN we should get the correct lane
    assert result == lane


def test_get_lane_from_sample_fastq_file_path_no_flowcell():
    # GIVEN a sample fastq path without a flow cell id
    lane = 4
    sample_fastq_path = Path(
        f"ACC12164A17_S367_L00{lane}_R1_001{FileExtensions.FASTQ}{FileExtensions.GZIP}",
    )

    # WHEN we get lane from the sample fastq file path
    result = get_lane_from_sample_fastq(sample_fastq_path)

    # THEN we should get the correct lane
    assert result == lane


def test_is_bcl2fastq_folder_structure(bcl2fastq_folder_structure: Path):
    # GIVEN a bcl2fastq folder structure
    # WHEN checking if it is a bcl2fastq folder structure
    is_bcl2fastq_folder_structure = is_bcl2fastq_demux_folder_structure(bcl2fastq_folder_structure)

    # THEN it should be a bcl2fastq folder structure
    assert is_bcl2fastq_folder_structure is True


def test_is_not_bcl2fastq_folder_structure(not_bcl2fastq_folder_structure: Path):
    # GIVEN not folder structure which is not formatted correctly for bcl2fastq

    # WHEN checking if it is a bcl2fastq folder structure
    is_bcl2fastq_folder_structure = is_bcl2fastq_demux_folder_structure(
        not_bcl2fastq_folder_structure
    )

    # THEN it should not be a bcl2fastq folder structure
    assert is_bcl2fastq_folder_structure is False


def test_get_bcl_converter_bcl2fastq(bcl2fastq_folder_structure: Path):
    # GIVEN a bcl2fastq folder structure
    # WHEN getting the bcl converter
    bcl_converter: str = get_bcl_converter_name(bcl2fastq_folder_structure)

    # THEN it should be bcl2fastq
    assert bcl_converter == BclConverter.BCL2FASTQ


def test_get_bcl_converter_bclconvert():
    # GIVEN any folder structure which is not bcl2fastq
    # WHEN getting the bcl converter
    bcl_converter: str = get_bcl_converter_name(Path())

    # THEN it returns bclconvert
    assert bcl_converter == BclConverter.BCLCONVERT


def test_validate_demux_complete_flow_cell_directory_when_it_exists(tmp_path: Path):
    # GIVEN a temporary directory as the flow cell directory with a demux complete file
    flow_cell_directory = tmp_path
    Path(flow_cell_directory, DemultiplexingDirsAndFiles.DEMUX_COMPLETE).touch()

    # WHEN the create_delivery_file_in_flow_cell_directory function is called
    create_delivery_file_in_flow_cell_directory(flow_cell_directory)

    # THEN a delivery file should exist in the flow cell directory
    assert (flow_cell_directory / DemultiplexingDirsAndFiles.DEMUX_COMPLETE).exists()


def test_validate_demux_complete_flow_cell_directory_when_it_does_not_exist(tmp_path: Path):
    # GIVEN a temporary directory as the flow cell directory without a demux complete file
    flow_cell_directory = tmp_path

    # WHEN the create_delivery_file_in_flow_cell_directory function is called
    create_delivery_file_in_flow_cell_directory(flow_cell_directory)

    # THEN a delivery file should not exist in the flow cell directory
    assert not (flow_cell_directory / DemultiplexingDirsAndFiles.DEMUX_COMPLETE).exists()


def test_get_q30_threshold():
    # GIVEN a specific sequencer type
    sequencer_type = Sequencers.HISEQGA

    # WHEN getting the Q30 threshold for the sequencer type
    q30_threshold = get_q30_threshold(sequencer_type)

    # THEN the correct Q30 threshold should be returned
    assert q30_threshold == FLOWCELL_Q30_THRESHOLD[sequencer_type]


def test_get_sample_sheet_path_found(tmp_path: Path):
    # GIVEN a temporary directory as the flow cell directory
    flow_cell_directory = tmp_path

    # GIVEN a sample sheet file in the flow cell directory
    sample_sheet_path = Path(flow_cell_directory, DemultiplexingDirsAndFiles.SAMPLE_SHEET_FILE_NAME)
    sample_sheet_path.touch()

    # WHEN the sample sheet is retrieved
    found_sample_sheet_path = get_sample_sheet_path(flow_cell_directory)

    # THEN the path to the sample sheet file should be returned
    assert found_sample_sheet_path == sample_sheet_path


def test_get_sample_sheet_path_found_in_nested_directory(tmp_path: Path):
    # GIVEN a temporary directory as the flow cell directory
    flow_cell_directory = tmp_path

    # GIVEN a nested directory within the flow cell directory
    nested_directory = Path(flow_cell_directory, "nested")
    nested_directory.mkdir()

    # GIVEN a sample sheet file in the nested directory
    sample_sheet_path = Path(nested_directory, DemultiplexingDirsAndFiles.SAMPLE_SHEET_FILE_NAME)
    sample_sheet_path.touch()

    # WHEN the sample sheet is retrieved
    found_sample_sheet_path = get_sample_sheet_path(flow_cell_directory)

    # THEN the path to the sample sheet file should be returned
    assert found_sample_sheet_path == sample_sheet_path


def test_get_sample_sheet_path_not_found(tmp_path: Path):
    # GIVEN a temporary directory as the flow cell directory without a sample sheet file
    flow_cell_directory = tmp_path

    # WHEN the sample sheet is retrieved
    # THEN a FileNotFoundError should be raised
    with pytest.raises(FileNotFoundError):
        get_sample_sheet_path(flow_cell_directory)


@patch("cg.meta.demultiplex.utils.is_flow_cell_directory_valid", return_value=False)
# GIVEN a flow cell directory which is not valid
# WHEN parsing the flow cell directory data
# THEN a FlowCellError should be raised
def test_parse_flow_cell_directory_data_invalid(mocked_function):
    with pytest.raises(FlowCellError):
        parse_flow_cell_directory_data(Path("dummy_path"), "dummy_bcl_converter")


@patch("cg.meta.demultiplex.utils.is_flow_cell_directory_valid", return_value=True)
def test_parse_flow_cell_directory_data_valid(mocked_function):
    # GIVEN a flow cell directory which is valid
    # WHEN parsing the flow cell directory data
    flow_cell_run_directory = "20230508_LH00188_0003_A22522YLT3"
    result = parse_flow_cell_directory_data(Path(flow_cell_run_directory), "dummy_bcl_converter")

    # THEN a FlowCellDirectoryData object should be returned
    assert isinstance(result, FlowCellDirectoryData)

    # THEN the flow cell path and bcl converter should be set
    assert result.path == Path(flow_cell_run_directory)
    assert result.bcl_converter == "dummy_bcl_converter"


def test_copy_sample_sheet(demultiplex_context: CGConfig):
    # GIVEN a DemuxPostProcessing API
    demux_post_processing_api = DemuxPostProcessingAPI(demultiplex_context)

    # GIVEN a sample sheet in the run directory
    sample_sheet_path = Path(
        demux_post_processing_api.demux_api.run_dir,
    )
    sample_sheet_path.mkdir(parents=True, exist_ok=True)
    Path(sample_sheet_path, DemultiplexingDirsAndFiles.SAMPLE_SHEET_FILE_NAME).touch()

    # GIVEN a sample sheet target path
    target_sample_sheet_path = Path(demux_post_processing_api.demux_api.out_dir)

    # WHEN copying the sample sheet
    copy_sample_sheet(
        sample_sheet_source_directory=sample_sheet_path,
        sample_sheet_destination_directory=target_sample_sheet_path,
    )

    # THEN the sample sheet was copied to the out directory
    assert Path(
        target_sample_sheet_path,
        DemultiplexingDirsAndFiles.SAMPLE_SHEET_FILE_NAME,
    ).exists()



import logging

from pathlib import Path
from typing import Any, List, Optional, Dict, Union

from pydantic.v1 import ValidationError

from cg.apps.mip.confighandler import ConfigHandler
from cg.constants import COLLABORATORS, COMBOS, GenePanelMasterList, Pipeline, FileExtensions
from cg.constants.constants import FileFormat
from cg.constants.housekeeper_tags import HkMipAnalysisTag
from cg.exc import CgError
from cg.io.controller import WriteFile, ReadFile
from cg.meta.workflow.analysis import AnalysisAPI
from cg.meta.workflow.fastq import MipFastqHandler
from cg.models.cg_config import CGConfig
from cg.models.mip.mip_analysis import MipAnalysis
from cg.models.mip.mip_config import MipBaseConfig
from cg.models.mip.mip_metrics_deliverables import MIPMetricsDeliverables
from cg.models.mip.mip_sample_info import MipBaseSampleInfo
from cg.store.models import BedVersion, FamilySample, Family, Sample

CLI_OPTIONS = {
    "config": {"option": "--config_file"},
    "slurm_quality_of_service": {"option": "--slurm_quality_of_service"},
    "email": {"option": "--email"},
    "base": {"option": "--cluster_constant_path"},
    "dryrun": {"option": "--dry_run_all"},
    "gene_list": {"option": "--vcfparser_slt_fl"},
    "max_gaussian": {"option": "--gatk_varrecal_snv_max_gau"},
    "skip_evaluation": {"option": "--qccollect_skip_evaluation"},
    "start_after": {"option": "--start_after_recipe"},
    "start_with": {"option": "--start_with_recipe"},
    "use_bwa_mem": {"option": "--bwa_mem 1 --bwa_mem2 0"},
}
LOG = logging.getLogger(__name__)


class MipAnalysisAPI(AnalysisAPI):
    """The workflow is accessed through Trailblazer but cg provides additional conventions and
    hooks into the status database that makes managing analyses simpler"""

    def __init__(self, config: CGConfig, pipeline: Pipeline):
        super().__init__(pipeline, config)

    @property
    def root(self) -> str:
        raise NotImplementedError

    @property
    def conda_env(self) -> str:
        raise NotImplementedError

    @property
    def mip_pipeline(self) -> str:
        raise NotImplementedError

    @property
    def script(self) -> str:
        raise NotImplementedError

    @property
    def fastq_handler(self):
        return MipFastqHandler

    def get_pedigree_config_path(self, case_id: str) -> Path:
        return Path(self.root, case_id, "pedigree.yaml")

    def get_case_config_path(self, case_id: str) -> Path:
        return Path(self.root, case_id, "analysis", f"{case_id}_config.yaml")

    def get_deliverables_file_path(self, case_id: str) -> Path:
        """
        Location in working directory where deliverables file will be stored upon completion of analysis.
        Deliverables file is used to communicate paths and tag definitions for files in a finished analysis
        """
        return Path(self.root, case_id, "analysis", f"{case_id}_deliverables.yaml")

    def get_sample_info_path(self, case_id: str) -> Path:
        """Get case analysis sample info path"""
        return Path(self.root, case_id, "analysis", f"{case_id}_qc_sample_info.yaml")

    def get_panel_bed(self, panel_bed: str = None) -> Optional[str]:
        """Check and return BED gene panel."""
        if not panel_bed:
            return None
        if panel_bed.endswith(FileExtensions.BED):
            return panel_bed
        bed_version: Optional[BedVersion] = self.status_db.get_bed_version_by_short_name(
            bed_version_short_name=panel_bed
        )
        if not bed_version:
            raise CgError("Please provide a valid panel shortname or a path to panel.bed file!")
        return bed_version.filename

    def pedigree_config(self, case_id: str, panel_bed: str = None) -> dict:
        """Make the MIP pedigree config. Meta data for the family is taken from the family object
        and converted to MIP format via trailblazer.
        """

        # Validate and reformat to MIP pedigree config format
        case_obj: Family = self.status_db.get_case_by_internal_id(internal_id=case_id)
        return ConfigHandler.make_pedigree_config(
            data={
                "case": case_obj.internal_id,
                "default_gene_panels": case_obj.panels,
                "samples": [
                    self.config_sample(link_obj=link_obj, panel_bed=panel_bed)
                    for link_obj in case_obj.links
                ],
            }
        )

    def write_pedigree_config(self, case_id: str, data: dict) -> None:
        """Write the pedigree config to the the case dir"""
        out_dir = self.get_case_path(case_id=case_id)
        out_dir.mkdir(parents=True, exist_ok=True)
        pedigree_config_path: Path = self.get_pedigree_config_path(case_id=case_id)
        WriteFile.write_file_from_content(
            content=data, file_format=FileFormat.YAML, file_path=pedigree_config_path
        )
        LOG.info("Config file saved to %s", pedigree_config_path)

    @staticmethod
    def get_sample_data(link_obj: FamilySample) -> Dict[str, Union[str, int]]:
        """Return sample specific data."""
        return {
            "sample_id": link_obj.sample.internal_id,
            "sample_display_name": link_obj.sample.name,
            "analysis_type": link_obj.sample.application_version.application.analysis_type,
            "sex": link_obj.sample.sex,
            "phenotype": link_obj.status,
            "expected_coverage": link_obj.sample.application_version.application.min_sequencing_depth,
        }

    def get_sample_fastq_destination_dir(self, case: Family, sample: Sample) -> Path:
        """Return the path to the FASTQ destination directory."""
        return Path(
            self.root,
            case.internal_id,
            sample.application_version.application.analysis_type,
            sample.internal_id,
            FileFormat.FASTQ,
        )

    def link_fastq_files(self, case_id: str, dry_run: bool = False) -> None:
        case_obj = self.status_db.get_case_by_internal_id(internal_id=case_id)
        for link in case_obj.links:
            self.link_fastq_files_for_sample(
                case_obj=case_obj,
                sample_obj=link.sample,
            )

    def write_panel(self, case_id: str, content: List[str]):
        """Write the gene panel to case dir"""
        out_dir = Path(self.root, case_id)
        out_dir.mkdir(parents=True, exist_ok=True)
        out_path = Path(out_dir, "gene_panels.bed")
        with out_path.open("w") as out_handle:
            out_handle.write("\n".join(content))

    @staticmethod
    def convert_panels(customer: str, default_panels: List[str]) -> List[str]:
        """Convert between default panels and all panels included in gene list."""
        # check if all default panels are part of master list
        master_list: List[str] = GenePanelMasterList.get_panel_names()
        if customer in COLLABORATORS and set(default_panels).issubset(master_list):
            return master_list

        # the rest are handled the same way
        all_panels = set(default_panels)

        # fill in extra panels if selection is part of a combo
        for panel in default_panels:
            if panel in COMBOS:
                for extra_panel in COMBOS[panel]:
                    all_panels.add(extra_panel)

        # add OMIM to every panel choice
        all_panels.add(GenePanelMasterList.OMIM_AUTO)

        return list(all_panels)

    def _get_latest_raw_file(self, family_id: str, tags: List[str]) -> Any:
        """Get a python object file for a tag and a family ."""

        last_version = self.housekeeper_api.last_version(bundle=family_id)

        analysis_files = self.housekeeper_api.files(
            bundle=family_id, version=last_version.id, tags=tags
        ).all()

        if analysis_files:
            analysis_file_raw = self._open_bundle_file(analysis_files[0].path)
        else:
            raise LOG.warning(
                "No post analysis files received from DeliverAPI for '%s'",
                family_id,
            )

        return analysis_file_raw

    def _open_bundle_file(self, relative_file_path: str) -> Any:
        """Open a bundle file and return it as an Python object"""

        full_file_path: Path = Path(self.housekeeper_api.get_root_dir()).joinpath(
            relative_file_path
        )
        return ReadFile.get_content_from_file(file_format=FileFormat.YAML, file_path=full_file_path)

    def get_latest_metadata(self, family_id: str) -> MipAnalysis:
        """Get the latest trending data for a family"""

        mip_config_raw = self._get_latest_raw_file(
            family_id=family_id, tags=HkMipAnalysisTag.CONFIG
        )
        qc_metrics_raw = self._get_latest_raw_file(
            family_id=family_id, tags=HkMipAnalysisTag.QC_METRICS
        )
        sample_info_raw = self._get_latest_raw_file(
            family_id=family_id, tags=HkMipAnalysisTag.SAMPLE_INFO
        )
        if mip_config_raw and qc_metrics_raw and sample_info_raw:
            try:
                mip_analysis: MipAnalysis = self.parse_analysis(
                    config_raw=mip_config_raw,
                    qc_metrics_raw=qc_metrics_raw,
                    sample_info_raw=sample_info_raw,
                )
                return mip_analysis
            except ValidationError as error:
                LOG.error(
                    "get_latest_metadata failed for '%s', missing attribute: %s",
                    family_id,
                    error,
                )
                raise error
        else:
            LOG.error(f"Unable to retrieve the latest metadata for {family_id}")
            raise CgError

    def parse_analysis(
        self, config_raw: dict, qc_metrics_raw: dict, sample_info_raw: dict
    ) -> MipAnalysis:
        """Parses the output analysis files from MIP

        Args:
            config_raw (dict): raw YAML input from MIP analysis config file
            qc_metrics_raw (dict): raw YAML input from MIP analysis qc metric file
            sample_info_raw (dict): raw YAML input from MIP analysis qc sample info file
        Returns:
            MipAnalysis: parsed MIP analysis data
        """
        mip_config: MipBaseConfig = MipBaseConfig(**config_raw)
        qc_metrics = MIPMetricsDeliverables(**qc_metrics_raw)
        sample_info: MipBaseSampleInfo = MipBaseSampleInfo(**sample_info_raw)

        return MipAnalysis(
            case=mip_config.case_id,
            genome_build=sample_info.genome_build,
            sample_id_metrics=qc_metrics.sample_id_metrics,
            mip_version=sample_info.mip_version,
            rank_model_version=sample_info.rank_model_version,
            sample_ids=mip_config.sample_ids,
            sv_rank_model_version=sample_info.sv_rank_model_version,
        )

    def get_skip_evaluation_flag(self, case_id: str, skip_evaluation: bool) -> bool:
        """If any sample in this case is downsampled or external, returns true"""
        if skip_evaluation:
            return True
        case_obj = self.status_db.get_case_by_internal_id(internal_id=case_id)
        for link_obj in case_obj.links:
            downsampled = isinstance(link_obj.sample.downsampled_to, int)
            external = link_obj.sample.application_version.application.is_external
            if downsampled or external:
                LOG.info(
                    "%s: downsampled/external - skip evaluation",
                    link_obj.sample.internal_id,
                )
                return True
        return False

    def get_cases_to_analyze(self) -> List[Family]:
        """Return cases to analyze."""
        cases_query: List[Family] = self.status_db.cases_to_analyze(
            pipeline=self.pipeline, threshold=self.threshold_reads
        )
        cases_to_analyze = []
        for case_obj in cases_query:
            if case_obj.action == "analyze" or not case_obj.latest_analyzed:
                cases_to_analyze.append(case_obj)
            elif (
                self.trailblazer_api.get_latest_analysis_status(case_id=case_obj.internal_id)
                == "failed"
            ):
                cases_to_analyze.append(case_obj)
        return cases_to_analyze

    @staticmethod
    def _append_value_for_non_flags(parameters: list, value) -> None:
        """Add the value of the non boolean options to the parameters"""
        if value is not True:
            parameters.append(value)

    @staticmethod
    def _cg_to_mip_option_map(parameters: list, mip_key) -> None:
        """Map cg options to MIP option syntax"""
        parameters.append(CLI_OPTIONS[mip_key]["option"])

    def run_analysis(self, case_id: str, command_args: dict, dry_run: bool) -> None:
        parameters = [
            case_id,
        ]
        for key, value in command_args.items():
            if value:
                self._cg_to_mip_option_map(parameters, key)
                self._append_value_for_non_flags(parameters, value)

        exit_code = self.process.run_command(dry_run=dry_run, parameters=parameters)
        for line in self.process.stdout_lines():
            LOG.info(line)
        for line in self.process.stderr_lines():
            LOG.info(line)
        return exit_code

    def get_case_path(self, case_id: str) -> Path:
        return Path(self.root, case_id)

    def get_trailblazer_config_path(self, case_id: str) -> Path:
        return Path(self.get_case_path(case_id=case_id), "analysis", "slurm_job_ids.yaml")

    def config_sample(self, link_obj: FamilySample, panel_bed: str) -> dict:
        raise NotImplementedError

    def get_pipeline_version(self, case_id: str) -> str:
        """Get MIP version from sample info file"""
        LOG.debug("Fetch pipeline version")
        sample_info_raw: dict = ReadFile.get_content_from_file(
            file_format=FileFormat.YAML, file_path=self.get_sample_info_path(case_id)
        )
        sample_info: MipBaseSampleInfo = MipBaseSampleInfo(**sample_info_raw)
        return sample_info.mip_version



from datetime import datetime
from pathlib import Path
from typing import List

from cg.apps.sequencing_metrics_parser.models.bcl2fastq_metrics import Bcl2FastqSampleLaneMetrics
from cg.apps.sequencing_metrics_parser.parsers.bcl2fastq import parse_bcl2fastq_sequencing_metrics
from cg.apps.sequencing_metrics_parser.sequencing_metrics_calculator import (
    average_quality_score,
    q30_ratio,
)
from cg.store.models import SampleLaneSequencingMetrics


def create_sample_lane_sequencing_metrics_from_bcl2fastq_for_flow_cell(
    flow_cell_dir: Path,
) -> List[SampleLaneSequencingMetrics]:
    """
    Parses the Bcl2fastq generated stats.json files and aggregates and calculates metrics for each sample in each lane.

    Args:
        flow_cell_dir (Path): Demultiplexed flow cell directory containing output from bcl2fastq

    Returns:
        List[SampleLaneSequencingMetrics]: A list of SampleLaneSequencingMetrics representing the sequencing
        metrics for each sample in each lane on the flow cell.
    """

    sample_lane_sequencing_metrics: List[SampleLaneSequencingMetrics] = []

    sample_and_lane_metrics: List[Bcl2FastqSampleLaneMetrics] = parse_bcl2fastq_sequencing_metrics(
        flow_cell_dir=flow_cell_dir
    )

    for raw_sample_metrics in sample_and_lane_metrics:
        metrics: SampleLaneSequencingMetrics = create_sample_lane_sequencing_metrics_from_bcl2fastq(
            bcl2fastq_sample_metrics=raw_sample_metrics
        )
        sample_lane_sequencing_metrics.append(metrics)

    return sample_lane_sequencing_metrics


def create_sample_lane_sequencing_metrics_from_bcl2fastq(
    bcl2fastq_sample_metrics: Bcl2FastqSampleLaneMetrics,
) -> SampleLaneSequencingMetrics:
    """
    Generates a SampleLaneSequencingMetrics based on the provided raw results.

    Args:
        raw_sample_metrics: Bcl2FastqSampleLaneMetrics: The raw sequencing metrics for a sample in a lane.

    Returns:
        SampleLaneSequencingMetrics: SampleLaneSequencingMetrics encapsulates the statistics for a sample
        in a lane on the flow cell.
    """

    sample_base_fraction_passing_q30: float = q30_ratio(
        q30_yield=bcl2fastq_sample_metrics.sample_total_yield_q30_in_lane,
        total_yield=bcl2fastq_sample_metrics.sample_total_yield_in_lane,
    )
    sample_base_mean_quality_score: float = average_quality_score(
        total_quality_score=bcl2fastq_sample_metrics.sample_total_quality_score_in_lane,
        total_yield=bcl2fastq_sample_metrics.sample_total_yield_in_lane,
    )

    return SampleLaneSequencingMetrics(
        flow_cell_name=bcl2fastq_sample_metrics.flow_cell_name,
        flow_cell_lane_number=bcl2fastq_sample_metrics.flow_cell_lane_number,
        sample_internal_id=bcl2fastq_sample_metrics.sample_id,
        sample_total_reads_in_lane=bcl2fastq_sample_metrics.sample_total_reads_in_lane,
        sample_base_fraction_passing_q30=sample_base_fraction_passing_q30,
        sample_base_mean_quality_score=sample_base_mean_quality_score,
        created_at=datetime.now(),
    )



"""Schemas for scout serialisation"""

from datetime import datetime
from typing import List, Optional, Dict

from pydantic.v1 import BaseModel, Field, validator
from typing_extensions import Literal

from cg.constants.gene_panel import GENOME_BUILD_37
from cg.constants.subject import (
    PlinkGender,
    Gender,
    PlinkPhenotypeStatus,
    RelationshipStatus,
)
from cg.constants.pedigree import Pedigree


class Individual(BaseModel):
    bam_file: Optional[str] = None
    individual_id: str
    sex: Literal[PlinkGender.UNKNOWN, PlinkGender.MALE, PlinkGender.FEMALE, Gender.OTHER]
    father: Optional[str]
    mother: Optional[str]
    phenotype: PlinkPhenotypeStatus
    analysis_type: str = "wgs"

    @validator(Pedigree.FATHER, Pedigree.MOTHER)
    def convert_to_zero(cls, value):
        if value is None:
            return RelationshipStatus.HAS_NO_PARENT
        return value

    @validator(Pedigree.SEX)
    def convert_sex_to_zero(cls, value):
        if value == Gender.OTHER:
            return PlinkGender.UNKNOWN
        return value


class Panel(BaseModel):
    panel_name: str


class Phenotype(BaseModel):
    phenotype_id: str
    feature: str


class Gene(BaseModel):
    hgnc_id: int
    hgnc_symbol: Optional[str]
    region_annotation: Optional[str]
    functional_annotation: Optional[str]
    sift_prediction: Optional[str]
    polyphen_prediction: Optional[str]


class DiagnosisPhenotypes(BaseModel):
    disease_nr: int
    disease_id: str
    description: str
    individuals: Optional[List[Dict[str, str]]]


class ScoutExportCase(BaseModel):
    id: str = Field(str, alias="_id")
    analysis_date: datetime
    owner: str
    causatives: Optional[List[str]] = None
    collaborators: List[str] = []
    individuals: List[Individual]
    genome_build: Optional[str]
    panels: Optional[List[Panel]]
    rank_model_version: Optional[str]
    sv_rank_model_version: Optional[str]
    rank_score_threshold: int = 5
    phenotype_terms: Optional[List[Phenotype]]
    phenotype_groups: Optional[List[Phenotype]]
    diagnosis_phenotypes: Optional[List[DiagnosisPhenotypes]]
    diagnosis_genes: Optional[List[int]]

    @validator("genome_build")
    def convert_genome_build(cls, value):
        if value is None:
            return GENOME_BUILD_37
        return value


class Genotype(BaseModel):
    genotype_call: str
    allele_depths: List[int]
    read_depth: int
    genotype_quality: int
    sample_id: str


class Variant(BaseModel):
    document_id: str
    variant_id: str
    chromosome: str
    position: int
    dbsnp_id: Optional[str] = None
    reference: str
    alternative: str
    quality: Optional[float] = None
    filters: Optional[List[str]] = None
    end: int
    rank_score: int
    category: str
    sub_category: str
    genes: Optional[List[Gene]]
    samples: List[Genotype]



"""CLI support to create config and/or start TAXPROFILER."""

import logging
from pathlib import Path

import click

from cg.cli.workflow.commands import ARGUMENT_CASE_ID
from cg.cli.workflow.taxprofiler.options import OPTION_INSTRUMENT_PLATFORM
from cg.constants.constants import MetaApis
from cg.constants.sequencing import SequencingPlatform
from cg.exc import CgError
from cg.meta.workflow.analysis import AnalysisAPI
from cg.meta.workflow.taxprofiler import TaxprofilerAnalysisAPI
from cg.models.cg_config import CGConfig


LOG = logging.getLogger(__name__)


@click.group(invoke_without_command=True)
@click.pass_context
def taxprofiler(context: click.Context) -> None:
    """nf-core/taxprofiler analysis workflow."""
    AnalysisAPI.get_help(context)
    context.obj.meta_apis[MetaApis.ANALYSIS_API] = TaxprofilerAnalysisAPI(
        config=context.obj,
    )


@taxprofiler.command("config-case")
@ARGUMENT_CASE_ID
@OPTION_INSTRUMENT_PLATFORM
@click.pass_obj
def config_case(context: CGConfig, case_id: str, instrument_platform: SequencingPlatform) -> None:
    """Create sample sheet file for Taxprofiler analysis for a given case_id."""
    analysis_api: TaxprofilerAnalysisAPI = context.meta_apis[MetaApis.ANALYSIS_API]

    try:
        analysis_api.status_db.verify_case_exists(case_internal_id=case_id)
        analysis_api.config_case(case_id=case_id, instrument_platform=instrument_platform, fasta="")

    except CgError as error:
        LOG.error(f"Could not create sample sheet: {error}")
        LOG.error(f"{case_id} could not be found in StatusDB!")
        raise click.Abort() from error



import os
import re
from pathlib import Path
from typing import List

from cg.apps.sequencing_metrics_parser.models.bcl2fastq_metrics import (
    Bcl2FastqSampleLaneMetrics,
    Bcl2FastqSampleLaneTileMetrics,
)
from cg.constants.demultiplexing import (
    BCL2FASTQ_METRICS_DIRECTORY_NAME,
    BCL2FASTQ_METRICS_FILE_NAME,
)


def parse_bcl2fastq_sequencing_metrics(
    flow_cell_dir: Path,
) -> List[Bcl2FastqSampleLaneMetrics]:
    """
    Parse stats.json files in specified Bcl2fastq demultiplex result directory.

    This function navigates through subdirectories in the given path, identifies
    and parses the Stats.json files from Bcl2fastq specifying metrics per sample, lane and tile
    on the flow cell and returns a list of parsed sequencing metrics resolved per tile.

    Parameters:
    demultiplex_result_directory (Path): Path to the demultiplexing results.

    Returns:
    List[Bcl2FastqSampleLaneMetrics]: List of parsed sequencing metrics per sample and lane.
    """
    tile_sequencing_metrics: List[
        Bcl2FastqSampleLaneTileMetrics
    ] = parse_bcl2fastq_raw_tile_metrics(demultiplex_result_directory=flow_cell_dir)

    sample_lane_sequencing_metrics: List[
        Bcl2FastqSampleLaneMetrics
    ] = aggregate_tile_metrics_per_sample_and_lane(tile_metrics=tile_sequencing_metrics)

    return sample_lane_sequencing_metrics


def parse_bcl2fastq_raw_tile_metrics(
    demultiplex_result_directory: Path,
) -> List[Bcl2FastqSampleLaneTileMetrics]:
    """
    Parse stats.json files in specified Bcl2fastq demultiplex result directory.

    This function navigates through subdirectories in the given path, identifies
    and parses the Stats.json files from Bcl2fastq specifying metrics per sample, lane and tile
    on the flow cell and returns a list of parsed sequencing metrics resolved per tile.

    Parameters:
    demultiplex_result_directory (Path): Path to the demultiplexing results.

    Returns:
    List[Bcl2FastqTileSequencingMetrics]: List of parsed sequencing metrics per tile.
    """

    tile_sequencing_metrics: List[Bcl2FastqSampleLaneTileMetrics] = []

    stats_json_paths: List[Path] = get_bcl2fastq_stats_paths(
        demultiplex_result_directory=demultiplex_result_directory
    )

    for stats_json_path in stats_json_paths:
        if not stats_json_path.exists():
            raise FileNotFoundError(f"File {stats_json_path} does not exist.")
        sequencing_metrics = Bcl2FastqSampleLaneTileMetrics.parse_file(stats_json_path)
        tile_sequencing_metrics.append(sequencing_metrics)

    return tile_sequencing_metrics


def aggregate_tile_metrics_per_sample_and_lane(
    tile_metrics: List[Bcl2FastqSampleLaneTileMetrics],
) -> List[Bcl2FastqSampleLaneMetrics]:
    """Aggregate the metrics parsed per sample and tile to be per sample and lane instead."""
    metrics = {}

    for tile_metric in tile_metrics:
        for conversion_result in tile_metric.conversion_results:
            for demux_result in conversion_result.demux_results:
                sample_lane_key = (
                    conversion_result.lane_number,
                    demux_result.sample_id,
                )

                sample_id: str = discard_index_sequence(sample_id_with_index=demux_result.sample_id)

                if sample_lane_key not in metrics:
                    metrics[sample_lane_key] = Bcl2FastqSampleLaneMetrics(
                        flow_cell_name=tile_metric.flow_cell_name,
                        flow_cell_lane_number=conversion_result.lane_number,
                        sample_id=sample_id,
                        sample_total_reads_in_lane=0,
                        sample_total_yield_in_lane=0,
                        sample_total_yield_q30_in_lane=0,
                        sample_total_quality_score_in_lane=0,
                    )
                # Double the total reads since they are reported in pairs
                metrics[sample_lane_key].sample_total_reads_in_lane += demux_result.number_reads * 2
                metrics[sample_lane_key].sample_total_yield_in_lane += demux_result.yield_
                metrics[sample_lane_key].sample_total_yield_q30_in_lane += sum(
                    [read_metric.yield_q30 for read_metric in demux_result.read_metrics]
                )
                metrics[sample_lane_key].sample_total_quality_score_in_lane += sum(
                    [read_metric.quality_score_sum for read_metric in demux_result.read_metrics]
                )

    return list(metrics.values())


def discard_index_sequence(sample_id_with_index: str) -> str:
    """Discard the index sequence from the sample id."""
    return sample_id_with_index.split("_")[0]


def get_bcl2fastq_stats_paths(demultiplex_result_directory: Path) -> List[Path]:
    """
    Identify and return paths to stats.json files in Bcl2fastq demultiplex result directory.

    This function looks through subdirectories in the given demultiplex directory,
    matching specific naming pattern (l<num>t<num>), and collects paths
    to any stats.json files found within a "Stats" subdirectory.

    Parameters:
    demultiplex_result_directory (Path): Path to the demultiplexing results.

    Returns:
    List[Path]: List of paths to identified stats.json files.
    """
    stats_json_paths = []
    pattern = re.compile(r"l\d+t\d+")

    for subdir in os.listdir(demultiplex_result_directory):
        if pattern.match(subdir):
            stats_json_path = (
                demultiplex_result_directory
                / subdir
                / BCL2FASTQ_METRICS_DIRECTORY_NAME
                / BCL2FASTQ_METRICS_FILE_NAME
            )
            if stats_json_path.is_file():
                stats_json_paths.append(stats_json_path)

    return stats_json_paths
